/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdlib.h>

#include "Common.h"


int memInit() {
	int i;

	memLUT = (long*)malloc(0x10000 * 4);
	memset(memLUT, 0, 0x10000 * 4);

	RAM = (char*)malloc(0x00200000);
	PPM = (char*)malloc(0x00010000);
	HWM = (char*)malloc(0x00010000);
	ROM = (char*)malloc(0x00080000);
	if (memLUT == NULL || RAM == NULL || PPM == NULL || HWM == NULL || ROM == NULL) {
		SysMessage("Error allocating memory"); return -1;
	}

	for (i=0; i<0x80; i++) memLUT[i + 0x0000] = (unsigned long)&RAM[(i & 0x1f) << 16];
	memcpy(memLUT + 0x8000, memLUT, 0x80 * 4);
	memcpy(memLUT + 0xa000, memLUT, 0x80 * 4);

	for (i=0; i<0x01; i++) memLUT[i + 0x1f00] = (unsigned long)&PPM[i << 16];

	for (i=0; i<0x01; i++) memLUT[i + 0x1f80] = (unsigned long)&HWM[i << 16];

	for (i=0; i<0x08; i++) memLUT[i + 0xbfc0] = (unsigned long)&ROM[i << 16];

	return 0;
}

void memReset() {
	FILE *f = NULL;
	char Bios[256];

	memset(RAM, 0, 0x00200000);
	memset(PPM, 0, 0x00010000);
	memset(HWM, 0, 0x00010000);

	if (strcmp(Config.Bios, "HLE")) {
		sprintf(Bios, "%s%s", Config.BiosDir, Config.Bios);
		f = fopen(Bios, "rb");

		if (f == NULL) {
			SysMessage ("Could not open bios:\"%s\". Enabling HLE Bios\n", Bios);
			memset(ROM, 0, 0x80000);
			Config.HLE = 1;
		}
		else {
			fread(ROM, 1, 0x80000, f);
			fclose(f);
			Config.HLE = 0;
		}
	} else Config.HLE = 1;
}

void memShutdown() {
	free(RAM);
	free(PPM);
	free(HWM);
	free(ROM);
	free(memLUT);
}

static int writeok=1;

unsigned char memRead8(unsigned long mem) {
	char *p;
	unsigned long t;

	t = mem >> 16;
	if (t == 0x1f80) {
		if (mem < 0x1f801000)
			return uHWM8(mem);
		else
			return hwRead8(mem);
	} else {
		p = (char *)(memLUT[t]);
		if (p != NULL) {
			return *(unsigned char *)(p + (mem & 0xffff));
		} else {
#ifdef CPU_LOG
			CPU_LOG("err lb %8.8lx\n", mem);
#endif
			return 0;
		}
	}
}

unsigned short memRead16(unsigned long mem) {
	char *p;
	unsigned long t;

	t = mem >> 16;
	if (t == 0x1f80) {
		if (mem < 0x1f801000)
			return uHWM16(mem);
		else
			return hwRead16(mem);
	} else {
		p = (char *)(memLUT[t]);
		if (p != NULL) {
			return *(unsigned short *)(p + (mem & 0xffff));
		} else {
#ifdef CPU_LOG
			CPU_LOG("err lh %8.8lx\n", mem);
#endif
			return 0;
		}
	}
}

unsigned long memRead32(unsigned long mem) {
	char *p;
	unsigned long t;

	t = mem >> 16;
	if (t == 0x1f80) {
		if (mem < 0x1f801000)
			return uHWM32(mem);
		else
			return hwRead32(mem);
	} else {
		p = (char *)(memLUT[t]);
		if (p != NULL) {
			return *(unsigned long *)(p + (mem & 0xffff));
		} else {
#ifdef CPU_LOG
			if (writeok) CPU_LOG("err lw %8.8lx\n", mem);
#endif
			return 0;
		}
	}
}


void memWrite8(unsigned long mem, unsigned char value) {
	char *p;
	unsigned long t;

	t = mem >> 16;
	if (t == 0x1f80) {
		if (mem < 0x1f801000)
			uHWM8(mem) = value;
		else
			hwWrite8(mem, value);
	} else {
		p = (char *)(memLUT[t]);
		if (p != NULL) {
			*(unsigned char *)(p + (mem & 0xffff)) = value;
			Cpu->Clear(mem, 1);
		} else {
#ifdef CPU_LOG
			CPU_LOG("err sb %8.8lx\n", mem);
#endif
		}
	}
}

void memWrite16(unsigned long mem, unsigned short value) {
	char *p;
	unsigned long t;

	t = mem >> 16;
	if (t == 0x1f80) {
		if (mem < 0x1f801000)
			uHWM16(mem) = value;
		else
			hwWrite16(mem, value);
	} else {
		p = (char *)(memLUT[t]);
		if (p != NULL) {
			*(unsigned short *)(p + (mem & 0xffff)) = value;
			Cpu->Clear(mem, 1);
		} else {
#ifdef CPU_LOG
			CPU_LOG("err sh %8.8lx\n", mem);
#endif
		}
	}
}

void memWrite32(unsigned long mem, unsigned long value) {
	char *p;
	unsigned long t;

	t = mem >> 16;
	if (t == 0x1f80) {
		if (mem < 0x1f801000)
			uHWM32(mem) = value;
		else
			hwWrite32(mem, value);
	} else {
		p = (char *)(memLUT[t]);
		if (p != NULL) {
			*(unsigned long *)(p + (mem & 0xffff)) = value;
			Cpu->Clear(mem, 1);
		} else {
			if (mem != 0xfffe0130) {
				if (!writeok) Cpu->Clear(mem, 1);
#ifdef CPU_LOG
				if (writeok) CPU_LOG("err sw %8.8lx\n", mem);
#endif
			} else {
				int i;

				switch (value) {
					case 0x800: case 0x804:
						if (writeok == 0) break;
						writeok = 0;
						memset(memLUT + 0x0000, 0, 0x80 * 4);
						memset(memLUT + 0x8000, 0, 0x80 * 4);
						memset(memLUT + 0xa000, 0, 0x80 * 4);
						break;
					case 0x1e988:
						if (writeok == 1) break;
						writeok = 1;
						for (i=0; i<0x80; i++) memLUT[i + 0x0000] = (unsigned long)&RAM[(i & 0x1f) << 16];
						memcpy(memLUT + 0x8000, memLUT, 0x80 * 4);
						memcpy(memLUT + 0xa000, memLUT, 0x80 * 4);
						break;
					default:
#ifdef CPU_LOG
						CPU_LOG("unk %8.8lx = %x\n", mem, value);
#endif
						break;
				}
			}
		}
	}
}

